#include <codegen.h>
#include <dlfcn.h>
#include <slibs/slibs.h>

static jit_state_t* _jit;
static void* handle;
static int reg = 0;

jit_reg_t new_reg() {
    return JIT_R(reg++);
}

CodegenResult codegen(const char* argv0, ASTNode node) {
    CodegenResult res;

    if(node.type == AST_PROGRAM) {
        init_jit(argv0);
        _jit = jit_new_state();
        handle = dlopen(NULL, RTLD_LAZY);
        if(!handle) {
            fprintf(stderr, "dlopen error: %s\n", dlerror());
            exit(EXIT_FAILURE);
        }

        jit_prolog();

        for(sl_vec_it(n, node.body)) {
            codegen(argv0, *n);
        }

        jit_epilog();

        void* func = jit_emit();
        jit_disassemble();

        res.func = func;
    }

    if(node.type == AST_CALL_EXPRESSION) {
        jit_prepare();

        reg = JIT_R0;
        for(size_t i = 0; i < node.params.size; i++) {
            jit_pushargr(codegen(argv0, node.params.data[i]).reg);
        }

        void* func = dlsym(handle, node.name);
        if(func == NULL) {
            fprintf(stderr, "dlsym error: %s\n", dlerror());
            exit(1);
        }

        jit_finishi(func);
    }

    if(node.type == AST_NUMBER_LITERAL) {
        int value;
        sscanf(node.value, "%d", &value);

        res.reg = new_reg();
        jit_movi(res.reg, value);
    }

    if(node.type == AST_STRING_LITERAL) {
        res.reg = new_reg();
        jit_movi(res.reg, (jit_word_t)node.value);
    }

    return res;
}
